#!/usr/bin/env python3

import threading
import socket
import enum
import queue
import readline
import signal
import sys

from time import sleep

PORT = 8912
KEY = 'lanchatv0.0.1'

QUIT_SIGNAL = threading.Event()

class MsgType(enum.Enum):
    BCAST = 0
    MSG = 1

def signal_handler(sig, frame):
        QUIT_SIGNAL.set()

class NetworkManager(threading.Thread):
    def __init__(self, username, sendq, recvq, port=8912):
        threading.Thread.__init__(self)
        self.username = username
        self.port = port
        self.sendq = sendq
        self.recvq = recvq


        self.recv_buf = ''

        # copied almost direct from 
        # https://stackoverflow.com/questions/21089268/python-service-discovery-advertise-a-service-across-a-local-network#21090815
        #create UDP socket
        self.bcast_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.bcast_s.bind(('', 0))
        #this is a broadcast socket
        self.bcast_s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        self.local_ip = self.get_ip()
        self.bcast_data = '{}:BCAST:{}:{}'.format(KEY, self.local_ip, self.username)

        self.recv_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) #create UDP socket
        self.recv_s.bind(('', self.port))
        self.recv_s.settimeout(1)

        print('Thread created on {}:{}'.format(self.local_ip, self.port))

    def run(self):
        print('Thread started')
        while not QUIT_SIGNAL.is_set():
            # broadcast
            self.bcast_s.sendto(self.bcast_data.encode('ASCII'), ('<broadcast>', self.port))

            # read messages
            while True:
                try:
                    # wait for a packet
                    data, addr = self.recv_s.recvfrom(1024)
                except socket.timeout as e:
                    break
                else:
                    valid, msgtype, ip, msg = self.parse_message(data)
                    if valid and ip != self.local_ip:
                        self.recvq.put((msgtype, ip, msg))

            # send messages
            while True:
                try:
                    to_send = self.sendq.get_nowait()
                except queue.Empty:
                    break
                else:
                    self.bcast_s.sendto('{}:MSG:{}:{}'.format(KEY, self.local_ip, to_send[1]).encode('ASCII'), (to_send[0], self.port))
                    print('Sent message')

    def parse_message(self, data):
        """Parse and validate a network message
        INPUT: data packet
        OUTPUT: (valid, msgtype, ip, msg)
        """

        fields = data.decode().split(':', 4)
        if len(fields) != 4:
            raise ValueError('Invalid Packet Format')

        if fields[0] != KEY:
            return (False, None, None, None)

        if fields[1] == 'BCAST':
            msgtype = MsgType.BCAST
        elif fields[1] == 'MSG':
            msgtype = MsgType.MSG

        ip = fields[2]
        message = fields[3]

        return (True, msgtype, ip, message)

    def get_ip(self):
        # shamelessly copied from 
        # https://stackoverflow.com/a/28950776
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            # should probably throw error
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

def main():
    signal.signal(signal.SIGINT, signal_handler)
    if len(sys.argv) != 2:
        print('Usage: {} <username>'.format(sys.argv[0]))
        return
    username = sys.argv[1]
    prompt = '\nOptions:\n - (q) Quit\n - (r) Refresh\n\nSend to:'
    neighbors = {}
    recv_q = queue.Queue()
    send_q = queue.Queue()
    manager = NetworkManager(username, send_q, recv_q, PORT)
    manager.start()
    while not QUIT_SIGNAL.is_set():
        # check the recieve queue
        while True:
            try:
                msgtype, ip, msg = recv_q.get(False)
            except queue.Empty:
                break
            else:
                if msgtype == MsgType.BCAST:
                    if ip not in neighbors:
                        neighbors[ip] = msg
                        print('New Neighbor! {} ({})'.format(msg, ip))
                if msgtype == MsgType.MSG:
                    username = ip
                    if neighbors[ip]:
                        username = neighbors[ip]
                    print(' - ({:>10}): {}'.format(username, msg))

        # prompt for input and send it
        print(prompt)
        iterator = 0
        for ip, username in neighbors.items():
            print(' - ({:>3}): {:10}'.format(iterator, username))
            iterator += 1

        choice = input('\n--> ')
        try:
            recipient = int(choice)
        except:
            recipient = -1

        if recipient in range(0, iterator):
            message = input('message for {} --> '.format(recipient))
            send_q.put((list(neighbors.keys())[recipient], message))
        elif choice.lower() == 'q':
            QUIT_SIGNAL.set()
            continue
        elif choice.lower() == 'r':
            continue
        else:
            print('Invalid input')
            continue

    QUIT_SIGNAL.set()
    manager.join()

if __name__== "__main__":
      main()
